@@ -2,15 +2,14 @@
 use crate::actors::connector;
 //use crate::actors::pool::{self, RillPoolTask};
 use anyhow::Error;
-//use async_trait::async_trait;
-//use futures::channel::mpsc;
+use async_trait::async_trait;
 use meio::Action;
-use rill_protocol::flow::core::{self, ActionEnvelope, TimedEvent};
+use rill_protocol::flow::core::{self, ActionEnvelope, FlowMode, TimedEvent};
 use rill_protocol::io::provider::{Description, Path, ProviderProtocol, Timestamp};
 use rill_protocol::io::transport::Direction;
 use std::sync::{Arc, Mutex, Weak};
 use std::time::{Duration, SystemTime};
-use tokio::sync::{mpsc, watch, Notify};
+use tokio::sync::mpsc;
 
 #[derive(Debug)]
 pub(crate) struct EventEnvelope<T: core::Flow> {
@@ -18,41 +17,49 @@ pub(crate) struct EventEnvelope<T: core::Flow> {
     pub event: T::Event,
 }
 
+pub(crate) enum ControlEvent<T> {
+    Flush,
+    AttachCallback { callback: BoxedCallback<T> },
+    // AttachCallbackSender { sender: ActionSender<T> },
+    DetachCallback,
+}
+
 impl<T: core::Flow> Action for EventEnvelope<T> {}
 
 // TODO: Remove that aliases and use raw types receivers in recorders.
 pub(crate) type DataSender<T> = mpsc::UnboundedSender<EventEnvelope<T>>;
 pub(crate) type DataReceiver<T> = mpsc::UnboundedReceiver<EventEnvelope<T>>;
 
-pub(crate) type ControlSender<T> = mpsc::UnboundedSender<ActionEnvelope<T>>;
+pub(crate) type ControlSender<T> = mpsc::UnboundedSender<ControlEvent<T>>;
+pub(crate) type ControlReceiver<T> = mpsc::UnboundedReceiver<ControlEvent<T>>;
 
-/// Watches for the control events.
-pub type Watcher<T> = mpsc::UnboundedReceiver<ActionEnvelope<T>>;
+/// A sender for actions wrapped with an envelope.
+pub type ActionSender<T> = mpsc::UnboundedSender<ActionEnvelope<T>>;
+/// A receiver for actions.
+pub type ActionReceiver<T> = mpsc::UnboundedReceiver<ActionEnvelope<T>>;
+
+/// Creates a new control channel.
+pub fn channel<T: core::Flow>() -> (ActionSender<T>, ActionReceiver<T>) {
+    mpsc::unbounded_channel()
+}
+
+pub(crate) struct TracerOperator<T: core::Flow> {
+    pub mode: TracerMode<T>,
+    pub control_rx: Option<ControlReceiver<T>>,
+}
 
 pub(crate) enum TracerMode<T: core::Flow> {
-    /* TODO: THE Idea to implement storage:
-     *
-     * Routed recorder shares the state and listens requests to forward them
-     * and return responses.
-     *
-     * Routed {
-     *   initial_state: T, - aka lazy state / bootstrap state
-     *   interactor: Address<?> or spawned routine, - to send requests there to update individual states
-     * },
-     */
     /// Real-time mode
     Push {
         state: T,
         receiver: Option<DataReceiver<T>>,
-        /// For sending events to the `Tracer` instance
-        control_sender: Option<ControlSender<T>>,
     },
     /// Pulling for intensive streams with high-load activities
     Pull {
+        // TODO: Replace with `Arc` since data channel used
+        // to detect Tracers's termination
         state: Weak<Mutex<T>>,
-        interval: Duration,
-        notifier: Arc<Notify>,
-        // TODO: It's also possible to support `control_sender`
+        interval: Option<Duration>,
     },
 }
 
@@ -68,10 +75,6 @@ enum InnerMode<T: core::Flow> {
     },
     Pull {
         state: Arc<Mutex<T>>,
-        /// For sending notifications about important state changes
-        /// IMPORTANT! No join with `Arc` of `state`, because
-        /// state needs `Weak` to detect closing.
-        notifier: Arc<Notify>,
     },
 }
 
@@ -82,9 +85,8 @@ impl<T: core::Flow> Clone for InnerMode<T> {
             Self::Push { sender } => Self::Push {
                 sender: sender.clone(),
             },
-            Self::Pull { state, notifier } => Self::Pull {
+            Self::Pull { state } => Self::Pull {
                 state: state.clone(),
-                notifier: notifier.clone(),
             },
         }
     }
@@ -94,17 +96,16 @@ impl<T: core::Flow> Clone for InnerMode<T> {
 /// for checking the activitiy status of the `Tracer`.
 #[derive(Debug)]
 pub struct Tracer<T: core::Flow> {
-    /// The receiver that used to activate/deactivate streams.
-    active: watch::Receiver<bool>,
     description: Arc<Description>,
+    control_tx: ControlSender<T>,
     mode: InnerMode<T>,
 }
 
 impl<T: core::Flow> Clone for Tracer<T> {
     fn clone(&self) -> Self {
         Self {
-            active: self.active.clone(),
             description: self.description.clone(),
+            control_tx: self.control_tx.clone(),
             mode: self.mode.clone(),
         }
     }
@@ -123,58 +124,54 @@ impl<T: core::Flow> Eq for Tracer<T> {}
 
 impl<T: core::Flow> Tracer<T> {
     /// Create a new `Tracer`
-    pub fn new(state: T, path: Path, pull_interval: Option<Duration>) -> Self {
-        if let Some(duration) = pull_interval {
-            Self::new_pull(state, path, duration)
-        } else {
-            Self::new_push(state, path).0
+    pub fn new(state: T, path: Path, mode: FlowMode) -> Self {
+        match mode {
+            FlowMode::Realtime => Self::new_push(state, path),
+            FlowMode::Throttle { ms } => {
+                Self::new_pull(state, path, Some(Duration::from_millis(ms)))
+            }
+            FlowMode::FlushOnly => Self::new_pull(state, path, None),
         }
     }
 
     /// Create a `Push` mode `Tracer`
-    pub fn new_push(state: T, path: Path) -> (Self, Watcher<T>) {
+    pub fn new_push(state: T, path: Path) -> Self {
         let (tx, rx) = mpsc::unbounded_channel();
-        let (control_tx, control_rx) = mpsc::unbounded_channel();
         let mode = TracerMode::Push {
             state,
             receiver: Some(rx),
-            control_sender: Some(control_tx),
         };
         let inner_mode = InnerMode::Push { sender: tx };
-        (Self::new_inner(path, inner_mode, mode), control_rx)
+        Self::new_inner(path, inner_mode, mode)
     }
 
     /// Create a `Pull` mode `Tracer`
-    pub fn new_pull(state: T, path: Path, interval: Duration) -> Self {
+    pub fn new_pull(state: T, path: Path, interval: Option<Duration>) -> Self {
         let state = Arc::new(Mutex::new(state));
-        let notifier = Arc::new(Notify::new());
         let mode = TracerMode::Pull {
             state: Arc::downgrade(&state),
             interval,
-            notifier: notifier.clone(),
         };
-        let inner_mode = InnerMode::Pull { state, notifier };
+        let inner_mode = InnerMode::Pull { state };
         Self::new_inner(path, inner_mode, mode)
     }
 
     fn new_inner(path: Path, inner_mode: InnerMode<T>, mode: TracerMode<T>) -> Self {
-        let stream_type = T::stream_type();
-        let info = format!("{} - {}", path, stream_type);
-        let description = Description {
-            path,
-            info,
-            stream_type,
+        let (control_tx, control_rx) = mpsc::unbounded_channel();
+        let operator = TracerOperator {
+            mode,
+            control_rx: Some(control_rx),
         };
-        // TODO: Remove this active watch channel?
-        let (_active_tx, active_rx) = watch::channel(true);
+        let stream_type = T::stream_type();
+        let description = Description { path, stream_type };
         log::trace!("Creating Tracer with path: {}", description.path);
         let description = Arc::new(description);
         let this = Tracer {
-            active: active_rx,
             description: description.clone(),
+            control_tx,
             mode: inner_mode,
         };
-        if let Err(err) = connector::DISTRIBUTOR.register_tracer(description, mode) {
+        if let Err(err) = connector::DISTRIBUTOR.register_tracer(description, operator) {
             log::error!(
                 "Can't register a Tracer. The worker can be terminated already: {}",
                 err
@@ -193,62 +190,41 @@ impl<T: core::Flow> Tracer<T> {
         &self.description
     }
 
+    /// Send an event to a `Recorder`.
+    // TODO: Consider using explicit direction value. What sould Broadcast be?
+    pub fn send(&self, event: T::Event, direction: Option<Direction<ProviderProtocol>>) {
+        match &self.mode {
+            InnerMode::Push { sender, .. } => {
+                let envelope = EventEnvelope { direction, event };
+                // And will never send an event
+                if let Err(err) = sender.send(envelope) {
+                    log::error!("Can't transfer data to sender of {}: {}", self.path(), err);
+                }
+            }
+            InnerMode::Pull { state, .. } => match state.lock() {
+                // `direction` ignored always in the `Pull` mode
+                Ok(ref mut state) => {
+                    T::apply(state, event);
+                }
+                Err(err) => {
+                    log::error!(
+                        "Can't lock the mutex to apply the changes of {}: {}",
+                        self.path(),
+                        err
+                    );
+                }
+            },
+        }
+    }
+
     /// Ask recorder to resend a state in the `Pull` mode.
     pub fn flush(&self) {
-        if self.is_active() {
-            match &self.mode {
-                InnerMode::Pull { notifier, .. } => {
-                    notifier.notify_one();
-                }
-                InnerMode::Push { .. } => {
-                    // TODO: Implement buffering and flushing.
-                    log::error!("Buffering and flushing is not supported in `Push` mode yet");
-                }
-            }
+        let event = ControlEvent::Flush;
+        if let Err(err) = self.control_tx.send(event) {
+            log::error!("Can't send a flush event to {}: {}", self.path(), err);
         }
     }
 
-    /// Send an event to a `Recorder`.
-    pub fn send(&self, event: T::Event, direction: Option<Direction<ProviderProtocol>>) {
-        if self.is_active() {
-            match &self.mode {
-                InnerMode::Push { sender, .. } => {
-                    let envelope = EventEnvelope { direction, event };
-                    // And will never send an event
-                    if let Err(err) = sender.send(envelope) {
-                        log::error!("Can't transfer data to sender: {}", err);
-                    }
-                }
-                InnerMode::Pull { state, .. } => match state.lock() {
-                    // `direction` ignored always in the `Pull` mode
-                    Ok(ref mut state) => {
-                        T::apply(state, event);
-                    }
-                    Err(err) => {
-                        log::error!(
-                            "Can't lock the mutex to apply the changes of {}: {}",
-                            self.path(),
-                            err
-                        );
-                    }
-                },
-            }
-        }
-    }
-
-    /*
-    /// Subscribe to the stream of the watcher.
-    pub fn subscribe(&mut self) -> Result<Watcher<T>, Error> {
-        match &mut self.mode {
-            InnerMode::Push { control_sender, .. } => Ok(control_sender.subscribe()),
-            InnerMode::Pull { .. } => {
-                log::error!("Can't receive state in pull mode of {}", self.path(),);
-                Err(Error::msg("Tracer::recv is not supported in pull mode."))
-            }
-        }
-    }
-    */
-
     /*
     /// Registers a callback to the flow.
     pub fn callback<F>(&mut self, func: F)
@@ -267,6 +243,82 @@ impl<T: core::Flow> Tracer<T> {
         }
     }
     */
+
+    /// Assign a callback
+    pub fn sync_callback<F>(&self, callback: F)
+    where
+        F: Fn(ActionEnvelope<T>) -> Result<(), Error>,
+        F: Send + Sync + 'static,
+    {
+        let sync_callback = SyncCallback::new(callback);
+        let callback = Box::new(sync_callback);
+        let event = ControlEvent::AttachCallback { callback };
+        if let Err(err) = self.control_tx.send(event) {
+            log::error!("Can't attach the callback from {}: {}", self.path(), err);
+        }
+    }
+
+    /// Removes the callback
+    pub fn detach_callback(&self) {
+        let event = ControlEvent::DetachCallback;
+        if let Err(err) = self.control_tx.send(event) {
+            log::error!("Can't detach the callback from {}: {}", self.path(), err);
+        }
+    }
+}
+
+/// The callback that called on flow's incoming actions.
+#[async_trait]
+pub trait ActionCallback<T: core::Flow>: Send + Sync + 'static {
+    /*
+    /// When at least one connection exists.
+    async fn awake(&mut self) {}
+
+    /// When all clients disconnected.
+    async fn suspend(&mut self) {}
+
+    /// A method to handle an action.
+    async fn handle_activity(
+        &mut self,
+        origin: ProviderReqId,
+        activity: Activity<T>,
+    ) -> Result<(), Error>;
+    */
+
+    /// A method to handle an action.
+    async fn handle_activity(&mut self, envelope: ActionEnvelope<T>) -> Result<(), Error>;
+}
+
+/// Boxed callback.
+pub type BoxedCallback<T> = Box<dyn ActionCallback<T>>;
+
+/// Sync callback.
+struct SyncCallback<F> {
+    func: Arc<F>,
+}
+
+impl<F> SyncCallback<F> {
+    fn new(func: F) -> Self {
+        Self {
+            func: Arc::new(func),
+        }
+    }
+}
+
+#[async_trait]
+impl<T, F> ActionCallback<T> for SyncCallback<F>
+where
+    T: core::Flow,
+    F: Fn(ActionEnvelope<T>) -> Result<(), Error>,
+    F: Send + Sync + 'static,
+{
+    async fn handle_activity(&mut self, envelope: ActionEnvelope<T>) -> Result<(), Error> {
+        let func = self.func.clone();
+        tokio::task::spawn_blocking(move || func(envelope))
+            .await
+            .map_err(Error::from)
+            .and_then(std::convert::identity)
+    }
 }
 
 /*
@@ -291,33 +343,6 @@ where
 }
 */
 
-impl<T: core::Flow> Tracer<T> {
-    /// Returns `true` is the `Tracer` has to send data.
-    pub fn is_active(&self) -> bool {
-        *self.active.borrow()
-    }
-
-    /* TODO: Remove or replace with an alternative
-    /// Use this method to detect when stream had activated.
-    ///
-    /// It's useful if you want to spawn async coroutine that
-    /// can read a batch of data, but will wait when some streams
-    /// will be activated to avoid resources wasting.
-    ///
-    /// When the generating coroutine active you can use `is_active`
-    /// method to detect when to change it to awaiting state again.
-    pub async fn when_activated(&mut self) -> Result<(), Error> {
-        loop {
-            if self.is_active() {
-                break;
-            }
-            self.active.changed().await?;
-        }
-        Ok(())
-    }
-    */
-}
-
 /// Wraps with timed event
 pub fn timed<T>(event: T) -> Option<TimedEvent<T>> {
     time_to_ts(None)
